// Multiple Lagged Fibonacci RNGs
// See Knuth, The Art of Computer Programming, v. 2, p. 26ff
#ifndef LFRNG_H
#define LFRNG_H

#include <limits>
#include <iostream>
#include <stdexcept>

namespace FPRNG {
class frng; // forward declaration for the sake of friendship
}

namespace LFRNG {

class rng {
    unsigned k, l;
    unsigned imk, iml;
    unsigned *hstr;

    struct pair {
        unsigned k, l;
        pair(unsigned i, unsigned j) : k(i), l(j) {}
    };

    const static unsigned n_lags = 3;
    const static pair lags[n_lags];

    void build(unsigned n);
    void random_init();
    void array_init(const unsigned *a);
    void copy_in(const rng& g);
protected:
    rng(const rng& g) { copy_in(g); }
    rng& operator= (const rng& g);
    unsigned draw();
public:
    explicit rng(unsigned n = 98) { build(n); random_init(); }
    rng(unsigned n, const unsigned *a) {
        build(n);
        if (n==k) array_init(a); 
        else throw std::invalid_argument("unsupported length");
    }
    ~rng();
    static unsigned max_len() { return lags[n_lags-1].k; }
    unsigned len() const { return k; }
    unsigned operator() () { return draw(); }
    friend std::ostream& operator<< (std::ostream& s, const rng& g);
    friend std::istream& operator>> (std::istream& s, rng& g);
    
    friend class FPRNG::frng;
};

class coin : public rng {
public:
    explicit coin(unsigned n=98) : rng(n) {}
    coin(unsigned n, const unsigned *a) : rng(n,a) {}
    bool operator() () {
             return rng::draw() > (std::numeric_limits<unsigned>::max()/2); }
};

class oddrng : public rng {
public:
    explicit oddrng(unsigned n=98) : rng(n) {}
    oddrng(unsigned n, const unsigned *a) : rng(n,a) {}
    unsigned operator() () { return rng::draw() | 1; }
};


class nofirst : public rng {
    unsigned first;
    bool takeit;
public:
    struct first_twice : public std::runtime_error {
        first_twice(const first_twice& e) : std::runtime_error(e) {}
        first_twice(const char *s) : std::runtime_error(s) {}
    };
    
    explicit nofirst(unsigned n=98) : rng(n), takeit(true) {}
    nofirst(unsigned n, const unsigned *a) : rng(n,a), takeit(true) {}
    unsigned operator() () { 
        unsigned next = rng::draw();
        if (takeit) {
            first = next;
            takeit = false;
        } else if (next == first)
            throw first_twice("first one occurred again");
        return next;
    }

    friend std::ostream& operator<< (std::ostream& s, const nofirst& g);
    friend std::istream& operator>> (std::istream& s, nofirst& g);
};


} //namespace LFRNG

#endif
