#ifndef INTERPOLATOR_H
#define INTERPOLATOR_H

#include <cstring>
#include <cmath>
#include <stdexcept>
#include <algorithm>

using namespace std;

// helper class to represent a data point
// sort() needs copy assignment and built-in array do not have one
template<class F> struct point { F x, y; };
    
template<class T> struct pwconstant {          // piecewise constant interpolation policy
    static T interpolate(int n, const point<T> table[], T x) {
        int i;  // we'll need the counter value AFTER the loop
        
        if (x < table[0].x) throw std::out_of_range("interpolator");
        
        for (i = 1; i < n; ++i)         // linear serch is inefficient!
            if (x < table[i].x) break;   // this is only a naive example
        
        if (i == n && x != table[n-1].x) throw std::out_of_range("interpolator");
        
        return table[i-1].y;
    }
};

template<class T, class I> class interpolator {
    int points;
    
    point<T> *table;

// predicate functor for find_if()    
    template<class F> class same_x {
        point<F>& n;
    public:
        same_x(point<F>& a) : n(a) {}
        bool operator() (const point<F>& a) const {
            return n.x == a.x;         
        }
    };

// template function for sort()    
    template<class F> static bool less_x(const point<F>& a, const point<F>& b) {
        return a.x < b.x;         
    }

public:
    interpolator() : points(0), table(NULL) { }
    interpolator(int n) : points(n), table(new point<T>[n]) { 
// To ease enforcement of unique x values on set-up, we initialize with NANs
// C++03 has no standard way to say NAN, so we use sqrt(-1).
// In C++0X, we'll say FP_NAN 
        for (int i = 0; i < n; ++i)
            table[i].x = table[i].y = sqrt(-1.0); 
    }
    
    interpolator(int n, const point<T> d[]) : points(n), table(new point<T>[n]) {
        memcpy(table, d, n*sizeof(point<T>));
        sort(&table[0], &table[n], less_x<T>);
        // check uniqueness of x values
        for(int i=1; i<points; ++i)
            if (table[i-1].x == table[i].x)
                throw std::invalid_argument("interpolator: duplicated x");        
    }
    
    interpolator(int n, const T x[], const T y[]) : points(n), table(new point<T>[n]) {
        for (int i = 0; i < n; ++i) {
            table[i].x = x[i];
            table[i].y = y[i];
        }
        sort(table, table+n, less_x<T>);
        // check uniqueness of x values
        for(int i=1; i<points; ++i)
            if (table[i-1].x == table[i].x)
                throw std::invalid_argument("interpolator: duplicated x");        
    }
    interpolator(const interpolator& p);
    
    ~interpolator() { delete[] table; }
    
    interpolator& operator= (const interpolator& p);

    int rows() const { return points; }
    
    T x(int n) const;   // getter 
    T y(int n) const;   // getter 
    void setxy(int n, T x, T y);    // setter
    
    T operator() (T x) const {
        if (!table) throw std::domain_error("uninitialized interpolator");
        return I::interpolate(points, table, x);
    }
};

template<class T, class I> interpolator<T, I>::interpolator(const interpolator<T, I>& p) : points(p.points), table(NULL) {
    if (points > 0) {
        table = new point<T>[points];
        memcpy(table, p.table, points*sizeof(point<T>));
    }
}

template<class T, class I> interpolator<T, I>& interpolator<T, I>::operator= (const interpolator<T, I>& p) {
    if (this == &p) return *this;
    points = p.points;
    delete[] table; table = NULL;
    if (points > 0) {
        table = new point<T>[points];
        memcpy(table, p.table, points*sizeof(point<T>));
    }
    return *this;
}

template<class T, class I> inline T interpolator<T, I>::x(int n) const {  // getter 
    if (!table) throw std::domain_error("uninitialized interpolator");
    if (n < 0 || n >= points) throw std::out_of_range("interpolator");
    return table[n].x;
}

template<class T, class I> inline T interpolator<T, I>::y(int n) const {  // getter 
    if (!table) throw std::domain_error("uninitialized interpolator");
    if (n < 0 || n >= points) throw std::out_of_range("interpolator");
    return table[n].y;
}

template<class T, class I> inline void interpolator<T, I>::setxy(int n, T x, T y) {  // setter 
    if (!table) throw std::domain_error("uninitialized interpolator");
    if (n < 0 || n >= points) throw std::out_of_range("interpolator");

    if (table[n].x == x)  // it's unique, by design
        table[n].y = y;
    else {               // check if x already exists in table
        point<T> tmp;
        tmp.x = x;
        tmp.y = y; 

        point<T> *found = find_if(&table[0], &table[points], same_x<T>(tmp));
        if (found != table+points)
            throw std::invalid_argument("interpolator: duplicated x");                
        table[n] = tmp;
        sort(table, table+points, less_x<T>);
    }
}

#endif
