#include <vector>
#include <list>
#include <map>
#include <algorithm>
#include <functional>
#include <iostream>
#include <typeinfo>


// Example program exercising STL features

using namespace std;

// helper template function to output map elements

template<class TF, class TS>
  ostream& operator<< (ostream& s, const pair<TF,TS>& p) {
    return s << "(" << p.first << ", " << p.second << ")";
}


// Generic function template to output containers, possible because we
// restrict to basic, common methods and don't modify the structure
template<class T> void putto(ostream& s, const T& v) {
    s << typeid(v).name() << ": ";
    if (v.empty())
        s << "Empty container!";
    else {
        typename T::const_iterator i;
        for(i=v.begin(); i != v.end(); ++i)
            s << *i << ' ';
    }
    s << endl;
}


// Overload of the former for built-in arrays
template<class T> void putto(ostream& s, const T *b, const T *e) {
    s << typeid(b).name() << ": ";
    if (b == e)
        s << "Empty container!";
    else {
        const T *i;
        for(i=b; i != e; ++i)
            s << *i << ' ';
    }
    s << endl;
}


// predicate class to sort in incrising order, odd numbers before even ones
template<class T> struct oddFirst{
    bool operator() (const T& a, const T& b) {
        if ( (a&1)^(b&1) )
            return a&1;
        return a < b;
    }
};
    

int main() {
    int a[90];
    vector<int> v;
    list<int> l;
    map<int,bool> m;
    
    cout << boolalpha; // we want 'true' and 'false' instead of 1 and 0
    
    cout << "Initial state:" << endl;
    
    putto(cout, a+0, a+90);
    putto(cout, v);
    putto(cout, l);
    putto(cout, m);
    
    cout << endl << "Fill up:" << endl;
    
    // filling different containers
    for(int i=0; i<90; ++i) {
        a[i] = i+1;
        v.push_back(i+1);
        cerr << "Vector size: " << v.size() 
             << ", capacity: " << v.capacity() << endl;
        l.push_front(90-i);
        m.insert(make_pair(i+1, false));
    }
        
    putto(cout, a+0, a+90);
    putto(cout, v);
    putto(cout, l);
    putto(cout, m);

    cout << endl << "Removal:" << endl;
    
    // different ways to remove/erase
    for(int i=0; i<90; i += 11) {
        remove(a, a+90, i);
        v.erase(remove(v.begin(), v.end(), i), v.end());
        l.remove(i);
        m.erase(i);
    }
    cerr << "Vector size: " << v.size() 
         << ", capacity: " << v.capacity() << endl;
        
    putto(cout, a+0, a+90);
    putto(cout, v);
    putto(cout, l);
    putto(cout, m);

    cout << endl << "Refill:" << endl;

    // Let's refill in scrambled order to test sort() algorithm

    // v.clear(); // Will empty v without releasing the memory chunk
                  // All access via [] would be unchecked out-of-bounds accesses
                  // to the existing memory chunk, which is already big enough
                  // They apparently 'work', while v.empty() holds true
    // cerr << "Vector size: " << v.size() 
    //      << ", capacity: " << v.capacity() << endl;
    l.clear();
    for(int i=0; i<90; i += 2) {
        v[i] = a[i] = i+1;
        v[89-i] = a[89-i] = i+2;
        l.push_front(i+1);
        l.push_back(i+2);
    }
        
    putto(cout, a+0, a+90);
    putto(cout, v);
    putto(cout, l);

    cout << endl << "Sort:" << endl;
    
    sort(a+0, a+90);    
    sort(v.begin(), v.end());    
    l.sort(greater<int>());  // for decreasing order, default is increasing order  

    putto(cout, a+0, a+90);
    putto(cout, v);
    putto(cout, l);

    cout << endl << "Funny sort:" << endl;
    
    sort(a+0, a+90, oddFirst<int>());    
    sort(v.begin(), v.end(), oddFirst<int>());    
    l.sort(oddFirst<int>());    

    putto(cout, a+0, a+90);
    putto(cout, v);
    putto(cout, l);

    cout << endl << "Find:" << endl;
    
    // lists ARE different: no random-access iterators!
    int *a88 = find(a+0, a+90, 88);    
    vector<int>::const_iterator v88 = find(v.begin(), v.end(), 88);    
    list<int>::const_iterator l88 = find(l.begin(), l.end(), 88);    

    cout << "... " << *(a88-1) << " " << *a88 << " " << *(a88+1) << " ..." << endl;
    cout << "... " << *(v88-1) << " " << *v88 << " " << *(v88+1) << " ..." << endl;
    advance(l88, -1);
    cout << "... " << *l88;
    advance(l88, 1);
    cout << " " << *l88;
    cout << " " << *++l88 << " ..." << endl;

    return 0;
}